import math
from typing import List, Any

from . import *


class Collection(TopLevel):
    """The Collection class is a class that groups together a set of
    TopLevel objects that have something in common.

    Some examples of Collection objects:

    * Results of a query to find all Component objects in a repository
      that function as promoters.
    * A set of Component objects representing a library of genetic
      logic gates.
    * A "parts list" for Component with a complex design, containing
      both that component and all of the Component, Sequence, and
      Model objects used to provide its full specification.

    """

    def __init__(self, identity: str,
                 *, members: List[str] = None,
                 namespace: str = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 type_uri: str = SBOL_COLLECTION) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         namespace=namespace,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)
        self.members = ReferencedObject(self, SBOL_MEMBER, 0, math.inf,
                                        initial_value=members)

    def accept(self, visitor: Any) -> Any:
        """Invokes `visit_collection` on `visitor` with `self` as the only
        argument.

        :param visitor: The visitor instance
        :type visitor: Any
        :raises AttributeError: If visitor lacks a visit_collection method
        :return: Whatever `visitor.visit_collection` returns
        :rtype: Any

        """
        visitor.visit_collection(self)


class Experiment(Collection):
    """The purpose of the Experiment class is to aggregate
    ExperimentalData objects for subsequent analysis, usually in
    accordance with an experimental design. Namely, the member
    properties of an Experiment MUST refer to ExperimentalData
    objects.

    """

    def __init__(self, identity: str,
                 *, members: List[str] = None,
                 namespace: str = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None,
                 type_uri: str = SBOL_EXPERIMENT) -> None:
        super().__init__(identity=identity, type_uri=type_uri,
                         members=members, namespace=namespace,
                         attachments=attachments, name=name,
                         description=description, derived_from=derived_from,
                         generated_by=generated_by, measures=measures)

    def accept(self, visitor: Any) -> Any:
        """Invokes `visit_experiment` on `visitor` with `self` as the only
        argument.

        :param visitor: The visitor instance
        :type visitor: Any
        :raises AttributeError: If visitor lacks a visit_experiment method
        :return: Whatever `visitor.visit_experiment` returns
        :rtype: Any

        """
        visitor.visit_experiment(self)


Document.register_builder(SBOL_COLLECTION, Collection)
Document.register_builder(SBOL_EXPERIMENT, Experiment)



import posixpath
import unittest
import uuid

import rdflib

import sbol3


class TestCollection(unittest.TestCase):

    def setUp(self) -> None:
        sbol3.set_defaults()

    def tearDown(self) -> None:
        sbol3.set_defaults()

    def test_create(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        collection = sbol3.Collection('collection1')
        self.assertIsNotNone(collection)
        self.assertEqual(0, len(collection.members))
        self.assertEqual(sbol3.SBOL_COLLECTION, collection.type_uri)

    def test_read(self):
        identity = 'https://github.com/synbiodex/pysbol3/collection1'
        nt_data = f'<{identity}> <{rdflib.RDF.type}> <{sbol3.SBOL_COLLECTION}> .'
        doc = sbol3.Document()
        doc.read_string(nt_data, 'ttl')
        collection = doc.find(identity)
        self.assertIsNotNone(collection)
        self.assertIsInstance(collection, sbol3.Collection)

    def test_member_property(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        self.assertTrue(hasattr(sbol3, 'SBOL_MEMBER'))
        collection = sbol3.Collection('collection1')
        self.assertIn(sbol3.SBOL_MEMBER, collection._properties)
        self.assertNotIn(sbol3.SBOL_ORIENTATION, collection._properties)
        uris = ['https://github.com/synbiodex/pysbol3/thing1',
                'https://github.com/synbiodex/pysbol3/thing2']
        collection.members = uris
        self.assertIn(sbol3.SBOL_MEMBER, collection._properties)
        self.assertNotIn(sbol3.SBOL_ORIENTATION, collection._properties)
        self.assertEqual(uris, collection.members)

    # Namespace testing
    def test_namespace_deduced(self):
        namespace = 'https://github.com/synbiodex/pysbol3'
        identity = posixpath.join(namespace, 'collection1')
        collection = sbol3.Collection(identity)
        self.assertEqual(namespace, collection.namespace)
        # Now test a namespace with a trailing #
        identity2 = f'{namespace}#collection2'
        collection2 = sbol3.Collection(identity2)
        self.assertEqual(namespace, collection2.namespace)

    def test_namespace_set(self):
        # Test that namespace is properly set on an object after
        # using set_namespace()
        namespace = 'https://github.com/synbiodex/pysbol3'
        sbol3.set_namespace(namespace)
        collection = sbol3.Collection('collection1')
        self.assertEqual(namespace, collection.namespace)

    def test_namespace_none(self):
        # Test the exception case when a namespace cannot be deduced
        identity = uuid.uuid4().urn
        collection = sbol3.Collection(identity)
        # The namespace should be None in this case.
        # We can't determine a namespace from a UUID and we don't
        # want to error and break file loading if we can't
        # determine a namespace.
        self.assertIsNone(collection.namespace)


class TestExperiment(unittest.TestCase):

    def setUp(self) -> None:
        sbol3.set_defaults()

    def tearDown(self) -> None:
        sbol3.set_defaults()

    def test_create(self):
        sbol3.set_namespace('https://github.com/synbiodex/pysbol3')
        experiment = sbol3.Experiment('experiment1')
        self.assertIsNotNone(experiment)
        self.assertEqual(0, len(experiment.members))
        self.assertEqual(sbol3.SBOL_EXPERIMENT, experiment.type_uri)

    def test_read(self):
        identity = 'https://github.com/synbiodex/pysbol3/experiment1'
        nt_data = f'<{identity}> <{rdflib.RDF.type}> <{sbol3.SBOL_EXPERIMENT}> .'
        doc = sbol3.Document()
        doc.read_string(nt_data, 'ttl')
        experiment = doc.find(identity)
        self.assertIsNotNone(experiment)
        self.assertIsInstance(experiment, sbol3.Collection)


if __name__ == '__main__':
    unittest.main()



import unittest

import sbol3


class TestConfig(unittest.TestCase):

    # I wanted to have this in a base class but I couldn't get the
    # imports right. I kept getting "attempted relative import with
    # no known parent package".
    def setUp(self) -> None:
        sbol3.set_defaults()

    def tearDown(self) -> None:
        sbol3.set_defaults()

    def test_namespace(self):
        base_uri = 'https://github.com/synbiodex/pysbol3'
        sbol3.set_namespace(base_uri)
        self.assertEqual(base_uri, sbol3.get_namespace())

        # Example from SBOL 3.0 Section 5.1 (page 12)
        # See issue #80
        base_uri = 'https://synbiohub.org'
        sbol3.set_namespace(base_uri)
        self.assertEqual(base_uri, sbol3.get_namespace())

    def test_no_namespace(self):
        # This test requires that no objects have been created without
        # a namespace set. The warning only triggers on the first such
        # creation. There is probably a way to reset the triggering of
        # warnings if we need that in the future.
        #
        # Make sure there is no default namespace
        self.assertEqual(None, sbol3.get_namespace())
        # Make sure creation without a namespace generates a warning
        display_id = 'c1'
        with self.assertWarns(UserWarning) as cm:
            c = sbol3.Component(display_id, sbol3.SBO_DNA)
        # We expect the warning to come from object.py. Update this if
        # the logic changes.
        self.assertIn('object.py', cm.filename)
        # We expect the default namespace to appear in the identity
        self.assertIn(sbol3.PYSBOL3_DEFAULT_NAMESPACE, c.identity)
        # We expect the default namespace at the beginning of the identity
        self.assertTrue(c.identity.startswith(sbol3.PYSBOL3_DEFAULT_NAMESPACE))
        # We expect the display_id to appear in the identity
        self.assertIn(display_id, c.identity)
        # We expect the display_id to be at the end of the identity
        self.assertTrue(c.identity.endswith(display_id))


if __name__ == '__main__':
    unittest.main()



import copy
import math
import posixpath
from typing import List, Dict, Callable, Union
from urllib.parse import urlparse

from . import *


class TopLevel(Identified):
    """TopLevel is an abstract class that is extended by any Identified
    class that can be found at the top level of an SBOL document or
    file. In other words, TopLevel objects are not nested inside any
    other object via composite aggregation. Instead of nesting,
    composite TopLevel objects refer to subordinate TopLevel objects
    by their URIs using shared aggregation.

    """

    def __init__(self, identity: str, type_uri: str,
                 *, namespace: str = None,
                 attachments: List[str] = None,
                 name: str = None, description: str = None,
                 derived_from: List[str] = None,
                 generated_by: List[str] = None,
                 measures: List[SBOLObject] = None) -> None:
        # Check identity, which is required for a TopLevel
        # More checking on identity happens in Identified, but Identified
        # does not require an identity, only TopLevel does.
        if not identity or not isinstance(identity, str):
            raise ValueError('Identity must be a non-empty string')
        super().__init__(identity=identity, type_uri=type_uri,
                         name=name, description=description,
                         derived_from=derived_from, generated_by=generated_by,
                         measures=measures)
        if namespace is None:
            namespace = TopLevel.default_namespace(namespace, identity)
        self.namespace = URIProperty(self, SBOL_NAMESPACE, 1, 1,
                                     initial_value=namespace)
        self.attachments = ReferencedObject(self, SBOL_HAS_ATTACHMENT, 0, math.inf,
                                            initial_value=attachments)

    @staticmethod
    def default_namespace(namespace: Union[None, str], identity: str) -> str:
        # short circuit if a namespace is set
        if namespace is not None:
            return namespace
        # Try using the default namespace if set
        namespace = get_namespace()
        if namespace is None:
            # No default namespace was defined
            # Try parsing the identity as a URL
            parsed = urlparse(identity)
            if parsed.scheme and parsed.netloc and parsed.path:
                # This is a URL
                # Reverse index to drop the displayId and use the rest
                # for the namespace
                delim = posixpath.sep
                if '#' in identity:
                    delim = '#'
                namespace = identity[:identity.rindex(delim)]
        if namespace is None:
            # TODO: what should we do here? We haven't been able to determine
            #       a namespace. But in the case where we're loading a file
            #       that's probably ok. We don't want that loading to fail.
            pass
        return namespace

    def validate_identity(self, report: ValidationReport) -> None:
        # TODO: See section 5.1 for rules about identity for TopLevel
        pass

    def validate(self, report: ValidationReport = None) -> ValidationReport:
        report = super().validate(report)
        return report

    def _reparent_child_objects(self) -> None:
        for k, v in self.__dict__.items():
            if not isinstance(v, Property):
                continue
            if v.property_uri not in self._owned_objects:
                continue
            old_list = self._owned_objects[v.property_uri]
            if len(old_list) == 0:
                continue
            self._owned_objects[v.property_uri] = []
            if isinstance(v, ListProperty):
                setattr(self, k, old_list)
            else:
                setattr(self, k, old_list[0])

    def clone(self, new_identity: str) -> 'TopLevel':
        obj = copy.deepcopy(self)
        identity_map = {self.identity: obj}
        # Set identity of new object
        obj._identity = obj._make_identity(new_identity)
        # Set display_id of new object
        obj._display_id = obj._extract_display_id(obj._identity)
        # Drop the document pointer
        obj.document = None

        # Erase the identity and display_id of the entire tree
        obj.traverse(make_erase_identity_traverser(identity_map))

        # Now remove and re-add all child objects to update
        # their identities
        obj._reparent_child_objects()

        # TODO: Now remap any properties that reference old
        #       identities in the identity_map
        obj.traverse(make_update_references_traverser(identity_map))
        return obj


def make_erase_identity_traverser(identity_map: Dict[str, Identified])\
        -> Callable[[Identified], None]:
    def erase_identity_traverser(x):
        if isinstance(x, TopLevel):
            return
        identity_map[x.identity] = x
        x._identity = None
        x._display_id = None
        x.document = None
    return erase_identity_traverser


def make_update_references_traverser(identity_map: Dict[str, Identified])\
        -> Callable[[Identified], None]:
    def update_references_traverser(x):
        # Use the identity map to update references.
        # References to objects outside of the object
        # being cloned will be left as is.
        for k, v in x.__dict__.items():
            if not isinstance(v, Property):
                continue
            if v.property_uri not in x._properties:
                continue
            items = x._properties[v.property_uri]
            for i in range(len(items)):
                str_item = str(items[i])
                if str_item in identity_map:
                    new_reference = identity_map[str_item].identity
                    # The item is probably an rdflib.URIRef. We take
                    # the type of the item and use that type to
                    # construct a new instance of the same type.
                    # Hacky? yes, and it seems to work
                    constructor = type(items[i])
                    items[i] = constructor(new_reference)
    return update_references_traverser
